************************************************************************
        subroutine init_4f(n,gen12,gen34)
************************************************************************
*	select final state for e+e- --> WW --> 4 f in DPA
*	  n = 1:    n_l   l+     l'- anti-n_l'
*	  n = 2:    n_l   l+     d   anti-u
*	  n = 3:     u  anti-d   l-  anti-n_l
*	  n = 4:     u  anti-d   d'  anti-u'
*	            ----v-----   ------v------
*	generations:  gen12         gen34
*-----------------------------------------------------------------------
*       20.5.99 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	integer n,gen12,gen34,dir(7)

	common/fstate/qf(4),mf(4),nc4f,dir
	common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &		     mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &		     mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

C*** define particle directions
        dir(1) =  1
        dir(2) =  1
        dir(3) = -1
        dir(4) = -1
        dir(5) = -1
        dir(6) = -1

C*** leptonic, semileptonic, or hadronic final state
	if (n.eq.1) then
	  nc4f  =  1d0
	  qf(1) =  0d0
	  qf(2) = -1d0
	  qf(3) = -1d0
	  qf(4) =  0d0
	  mf(1) =  0d0
	  mf(2) =  ml(gen12)
	  mf(3) =  ml(gen34)
	  mf(4) =  0d0
	elseif (n.eq.2) then
	  nc4f  =  3d0
	  qf(1) =  0d0
	  qf(2) = -1d0
	  qf(3) = -1d0/3d0
	  qf(4) = +2d0/3d0
	  mf(1) =  0d0
	  mf(2) =  ml(gen12)
	  mf(3) =  mqm(gen34)
	  mf(4) =  mqp(gen34)
	elseif (n.eq.3) then
	  nc4f  =  3d0
	  qf(1) = +2d0/3d0
	  qf(2) = -1d0/3d0
	  qf(3) = -1d0
	  qf(4) =  0d0
	  mf(1) =  mqp(gen12)
	  mf(2) =  mqm(gen12)
	  mf(3) =  ml(gen34)
	  mf(4) =  0d0
	elseif (n.eq.4) then
	  nc4f  =  9d0
	  qf(1) = +2d0/3d0
	  qf(2) = -1d0/3d0
	  qf(3) = -1d0/3d0
	  qf(4) = +2d0/3d0
	  mf(1) =  mqp(gen12)
	  mf(2) =  mqm(gen12)
	  mf(3) =  mqm(gen34)
	  mf(4) =  mqp(gen34)
	else
	  write(*,*) 'No definite final state !'
	  stop
	endif

	end

************************************************************************
        function M2(p,pon)
************************************************************************
*	calculate |M|^2 for e+e- --> WW --> 4 f in DPA
*-----------------------------------------------------------------------
*	Options: 
*	    
*	    qalp  = 0:  alpha(0)  scheme	    
*	            1:  alpha(MZ) scheme	    
*	            2:         GF scheme	    
*	    
*	    qgw   = 0:  Gamma_W from experiment	    
*	            1:  Gamma_W from theory [ corrected in O(alpha) ]
*	    
*	    qbos  = 0:	no virtual bosonic RCs
*	            1:	virtual bosonic RCs with CONVENTIONAL FORMALISM
*		 	  [ see A.Denner, Fortschr. Phys. 41 (1993) 4 ]
*	            2:	virtual bosonic RCs with BACKGROUND-FIELD METHOD 
*			  [ see A.Denner, S.Dittmaier, G.Weiglein,
*					Nucl. Phys. B440 (1995) 95 ]
*
*	    qferm = 0:	no fermion-loop RCs
*	            1:	fermion-loop RCs included
*
*	    qqcd  = 0:	no QCD corrections
*	            1:	   QCD corrections included
*
*	    qreal = 0:	imaginary parts of one-loop integrals INcluded
*	            1:	imaginary parts of one-loop integrals EXcluded
*
*	    qborn = 0:  Born matrix element not included
*	            1:  DPA Born matrix element included
*	            2:  (off-shell) CC03 Born matrix element included
*
*	    qsoft = 0:	no contributions of soft-photon emission
*
*	    qsoft =1,2,3: include soft-photon factor including
*	                  only counterparts to virtual corrections 
*                         (no soft cutoff dependence !)
*                   1,2: corresponds to endpoint contribution 
*                        of dipole subtraction formalism:
*	            1:	conventions of S.Dittmaier, Nucl.Phys.B565(2000)69.
*	            2:	conventions of M.Roth, Dissertation, 
*					       ETH Zurich, No. 13363, 1999.
*                   3:  corresponds to YFS-inspired definition of
*                       finite virtual photon contribution  
*
*	      qnf = 0:  no non-factorizable corrections
*	            1:  non-fact. corrs. with on-shell  Coulomb singularity
*	            2:  non-fact. corrs. with off-shell Coulomb singularity
*				  times DPA Born cross section
*	            3:  non-fact. corrs. with off-shell Coulomb singularity
*				  times (off-shell) CC03 Born cross section
*
*	    qfast = 0:  formfactors for virtual corrections from full formulas
*	            1:  formfactors from expansion into Legendre polynomials
*
*	Auxiliary variables:
*	  mudim2  = reference scale mu^2 of dimensional regularization
*	  lambda  = photon mass appearing only as log(lambda)
*	  lambda2 = lambda^2
*	  deltas  = Delta E / Ebeam  with Delta E = soft-photon cut
*-----------------------------------------------------------------------
*       10.6.99 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 F(10,-1:1),cmw2,FWf1f2virt_QCD
	complex*16 mat(10,-1:1),matcmb(10,-1:1),matb(-1:1),fb(10,-1:1)
        real*8 p(7,0:3),pon(7,0:3),ps(7,0:3)
	integer dir(7)
	integer qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast
c options
        integer smc,sborn4,sborn5,sborng5,sisr,src,son,sqcdepem
        integer ssigepem4,ssigepem5,ssigepemg5,scoul5,sunw,spaw,satgc
c contrib
        integer qsigepem4,qsigepem5,qqcdepem
c processtype        
        integer type,nc,cc,sym,perm
        
	common/rcoptions/qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast
        common/fstate/qf(4),mf(4),nc4f,dir
	common/sme/mat,matcmb,matb,fb,mbcmb
	common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &		     mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &		     mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)
	common/mandel/s,t,u,kp2,km2
	common/offshell/kp2off,km2off,cmw2
        common/options/smc,sborn4,sborn5,sborng5,sisr,src,son,sqcdepem,
     $       ssigepem4,ssigepem5,ssigepemg5,scoul5,sunw,spaw,satgc
        common/contrib/qsigepem4,qsigepem5,qqcdepem
        common/processtype/type,nc,cc,sym,perm

	call calcsme(p,pon)

	mbcmboffsh = M2_Born_offsh(p,0,0)

	if (qfast.eq.0) then
	  call calcform(F)
	elseif (qfast.eq.1) then
	  call fastform(s,t,F)
	endif

	m2qcd = 0d0
	if (qqcd.eq.1.or.qqcd.eq.3) then
          FWp = FWf1f2virt_QCD(kp2off,qf(1),qf(2),mf(1),mf(2))
          FWm = FWf1f2virt_QCD(km2off,qf(3),qf(4),mf(3),mf(4))
c switch to convention for calculating the off-shell Born
c in the case of MC processes
         do i=1,6
            ps(i,0)=p(i,0)
            ps(i,1)=p(i,1)
            ps(i,2)=p(i,2)
            ps(i,3)=p(i,3)
         end do
         if(cc.eq.1.and.nc.eq.1.and.(sym.eq.0.or.sym.eq.3))then
            do i=0,3
               ps(4,i)=p(6,i)
               ps(6,i)=p(4,i)
            end do
         end if
         if(qqcd.eq.1) qsigepem4=1
         if(qqcd.eq.3) qsigepem4=5
         qqcdepem=0
         call setprods(ps,dir,6)
         m2b = M2_epem_4f(ps)
         m2qcd = 2d0*(FWp+FWm)*m2b
        endif

	m2soft = 0d0
	if (smc.ne.2.and.(qsoft.eq.1.or.qsoft.eq.2))
     $       m2soft = dsoft1(pon)*mbcmb
        if (smc.ne.2.and.(qqcd.eq.1.or.qqcd.eq.3)) 
     $       m2soft = m2soft + dsoftg(p)*m2b

	M2nonf = 0d0
	if ((qnf.eq.1).or.(qnf.eq.2)) then
	  M2nonf = dnonf(p,pon)*mbcmb
	elseif (qnf.eq.3) then
	  M2nonf = (dnonf(p,pon)-coulsing(s,kp2off,km2off,0))*mbcmb
     &		   + coulsing(s,kp2off,km2off,1)*mbcmboffsh
	endif

	M2 = 0d0
	if (qborn.eq.1) then
	  M2 = mbcmb
	elseif (qborn.eq.2) then
	  M2 = mbcmboffsh
	endif

	M2 = M2 + m2qcd + M2nonf + m2soft + dM2(F)

	end

************************************************************************
      	subroutine oftoon(p,pon,mw2,n)
************************************************************************
* 	projection of off-shell momenta to on-shell momenta for               
* 	e^+(p1)+e^-(p2) -> f1(p3)+ fbar2(p4) + f3(p5)+ fbar4(p6)              
* 	CMS of e^+e^- is assumed
*	different versions chosen by flag n:
*	  n = 1: directions of momenta of W^+, f1 and f3 are fixed
*         n = 2: directions of momenta of W^+, f2 and f3 are fixed
*	  n = 3: directions of momenta of W^+, f1 and f4 are fixed
*         n = 4: directions of momenta of W^+, f2 and f4 are fixed
*-----------------------------------------------------------------------
*  27.5.99  Stefan Dittmaier
************************************************************************
      	implicit real*8 (a-z)
      	real*8 p(7,0:3),pon(7,0:3),q34(0:3),q56(0:3)
      	integer i,n

	if (p(1,0).ne. p(2,0).or.p(1,1).ne.-p(2,1).or.
     &      p(1,2).ne.-p(2,2).or.p(1,3).ne.-p(2,3)) then
          write(*,*) 'oftoon: momenta not in CMS'
	  stop
	endif

c*** fixed angles of (p3+p4), p3, p5
	if (n.eq.1) then
          q34(1) = p(3,1) + p(4,1)
          q34(2) = p(3,2) + p(4,2)
          q34(3) = p(3,3) + p(4,3)
          n34    = sqrt(q34(1)*q34(1)+q34(2)*q34(2)+q34(3)*q34(3))
          r34    = p(1,0)*sqrt(1d0-mw2/(p(1,0)*p(1,0)))/n34
          q34(0) = p(1,0)
          q34(1) = q34(1)*r34
          q34(2) = q34(2)*r34
          q34(3) = q34(3)*r34
          q56(0) = p(1,0)
          q56(1) = -q34(1)
          q56(2) = -q34(2)
          q56(3) = -q34(3)
          e3     = mw2*p(3,0)*0.5d0
     *    /(q34(0)*p(3,0)-q34(1)*p(3,1)-q34(2)*p(3,2)-q34(3)*p(3,3))
          e5     = mw2*p(5,0)*0.5d0
     *    /(q56(0)*p(5,0)-q56(1)*p(5,1)-q56(2)*p(5,2)-q56(3)*p(5,3))
          do i=0,3
            pon(1,i)=p(1,i)
            pon(2,i)=p(2,i)
            pon(3,i)=p(3,i)/p(3,0)*E3
            pon(4,i)=q34(i)-pon(3,i)
            pon(5,i)=p(5,i)/p(5,0)*E5
            pon(6,i)=q56(i)-pon(5,i)
          enddo
	  return
c*** fixed angles of (p3+p4), p4, p5
	elseif (n.eq.2) then
          q34(1) = p(3,1) + p(4,1)
          q34(2) = p(3,2) + p(4,2)
          q34(3) = p(3,3) + p(4,3)
          n34    = sqrt(q34(1)*q34(1)+q34(2)*q34(2)+q34(3)*q34(3))
          r34    = p(1,0)*sqrt(1d0-mw2/(p(1,0)*p(1,0)))/n34
          q34(0) = p(1,0)
          q34(1) = q34(1)*r34
          q34(2) = q34(2)*r34
          q34(3) = q34(3)*r34
          q56(0) = p(1,0)
          q56(1) = -q34(1)
          q56(2) = -q34(2)
          q56(3) = -q34(3)
          e4     = mw2*p(4,0)*0.5d0
     *    /(q34(0)*p(4,0)-q34(1)*p(4,1)-q34(2)*p(4,2)-q34(3)*p(4,3))
          e5     = mw2*p(5,0)*0.5d0
     *    /(q56(0)*p(5,0)-q56(1)*p(5,1)-q56(2)*p(5,2)-q56(3)*p(5,3))
          do i=0,3
            pon(1,i)=p(1,i)
            pon(2,i)=p(2,i)
            pon(4,i)=p(4,i)/p(4,0)*E4
            pon(3,i)=q34(i)-pon(4,i)
            pon(5,i)=p(5,i)/p(5,0)*E5
            pon(6,i)=q56(i)-pon(5,i)
          enddo
	  return
c*** fixed angles of (p3+p4), p3, p6
	elseif (n.eq.3) then
          q34(1) = p(3,1) + p(4,1)
          q34(2) = p(3,2) + p(4,2)
          q34(3) = p(3,3) + p(4,3)
          n34    = sqrt(q34(1)*q34(1)+q34(2)*q34(2)+q34(3)*q34(3))
          r34    = p(1,0)*sqrt(1d0-mw2/(p(1,0)*p(1,0)))/n34
          q34(0) = p(1,0)
          q34(1) = q34(1)*r34
          q34(2) = q34(2)*r34
          q34(3) = q34(3)*r34
          q56(0) = p(1,0)
          q56(1) = -q34(1)
          q56(2) = -q34(2)
          q56(3) = -q34(3)
          e3     = mw2*p(3,0)*0.5d0
     *    /(q34(0)*p(3,0)-q34(1)*p(3,1)-q34(2)*p(3,2)-q34(3)*p(3,3))
          e6     = mw2*p(6,0)*0.5d0
     *    /(q56(0)*p(6,0)-q56(1)*p(6,1)-q56(2)*p(6,2)-q56(3)*p(6,3))
          do i=0,3
            pon(1,i)=p(1,i)
            pon(2,i)=p(2,i)
            pon(3,i)=p(3,i)/p(3,0)*E3
            pon(4,i)=q34(i)-pon(3,i)
            pon(6,i)=p(6,i)/p(6,0)*E6
            pon(5,i)=q56(i)-pon(6,i)
          enddo
	  return
c*** fixed angles of (p3+p4), p4, p6
	elseif (n.eq.4) then
          q34(1) = p(3,1) + p(4,1)
          q34(2) = p(3,2) + p(4,2)
          q34(3) = p(3,3) + p(4,3)
          n34    = sqrt(q34(1)*q34(1)+q34(2)*q34(2)+q34(3)*q34(3))
          r34    = p(1,0)*sqrt(1d0-mw2/(p(1,0)*p(1,0)))/n34
          q34(0) = p(1,0)
          q34(1) = q34(1)*r34
          q34(2) = q34(2)*r34
          q34(3) = q34(3)*r34
          q56(0) = p(1,0)
          q56(1) = -q34(1)
          q56(2) = -q34(2)
          q56(3) = -q34(3)
          e4     = mw2*p(4,0)*0.5d0
     *    /(q34(0)*p(4,0)-q34(1)*p(4,1)-q34(2)*p(4,2)-q34(3)*p(4,3))
          e6     = mw2*p(6,0)*0.5d0
     *    /(q56(0)*p(6,0)-q56(1)*p(6,1)-q56(2)*p(6,2)-q56(3)*p(6,3))
          do i=0,3
            pon(1,i)=p(1,i)
            pon(2,i)=p(2,i)
            pon(4,i)=p(4,i)/p(4,0)*E4
            pon(3,i)=q34(i)-pon(4,i)
            pon(6,i)=p(6,i)/p(6,0)*E6
            pon(5,i)=q56(i)-pon(6,i)
          enddo
	  return
      	endif

	write(*,*) 'On-shell projection not defined.'
	stop

      end

***********************************************************************
        subroutine fastform(s,t,F)
************************************************************************
*-----------------------------------------------------------------------
*       23.6.99 Stefan Dittmaier
************************************************************************
       	implicit real*8 (a-z)
        complex*16 F(10,-1:1),cf(10,-1:1,0:50)
	real*8 p(0:50)
	integer j,k,l,nl

        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)
        common/fourier/cf,nl

	beta = sqrt(1d0-4d0*mw2/s)
	cth  = (1d0-2d0*(mw2-t)/s)/beta

	do 10 j=1,10
	do 10 k=-1,1,2
	  F(j,k) = 0d0
10	continue
	
	call LegP(nl,cth,p)
	do 20 l=0,nl
	do 20 j=1,10
	do 20 k=-1,1,2
c*** form factors scaled by t !!!
	  F(j,k) = F(j,k) + cf(j,k,l)*p(l) / t
20	continue

	end

************************************************************************
      	subroutine Cfourier(ss,nnl,ng)
************************************************************************
*-----------------------------------------------------------------------
*       23.6.99 Stefan Dittmaier
************************************************************************
       	implicit real*8 (a-z)
        complex*16 F(10,-1:1),cf(10,-1:1,0:50)
      	REAL*8 XGN(20,20),WN(20,20),p(0:50)
      	INTEGER i,j,k,n,nl,nnl,l,NG
        integer qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast

        common/rcoptions/qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast
      	COMMON/STUETN/XGN,WN                                              
        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)
        common/mandel/s,t,u,kp2,km2
        common/fourier/cf,nl

	qfast = 0

	call stuetz

C calculate renormalization constants
        call calcrenc
        call calcrenc_bgf
        call calcrenc_ferm

	nl   = nnl
	s    = ss
	beta = sqrt(1d0-4d0*mw2/s)
	kp2  = mw2
	km2  = mw2

	do 30 l=0,nl
	do 30 j=1,10
	do 30 k=-1,1,2
	  cf(j,k,l) = 0d0
30	continue

      	do 10 n = 1,ng 
	do 10 i = -1,1,2
          cth = i*XGN(NG,n)
	  t   = mw2-s/2d0*(1d0-beta*cth)
	  u   = 2d0*mw2-s-t
          call calcform(F)
	  call LegP(nl,cth,p)
	  do 20 j=1,10
	  do 20 k=-1,1,2
	  do 20 l=0,nl
c*** form factors scaled by t !!!
	    cf(j,k,l) = cf(j,k,l) 
     &			+ WN(NG,n)*F(j,k)*p(l)*(2d0*l+1d0)/2d0 * t
20	  continue
10	continue

      end

***********************************************************************
        subroutine LegP(nl,x,p)
************************************************************************
*	Legendre polynomials
*-----------------------------------------------------------------------
*       23.6.99 Stefan Dittmaier
************************************************************************
       	implicit real*8 (a-z)
	real*8  p(0:50)
	integer i,nl

	if (abs(x).le.0.9d0) then

        P(0) = 1D0
        P(1) = X
        P(2) = ((-1D0+3D0*X**2))/2D0
        P(3) = (X*(-3D0+5D0*X**2))/2D0
        P(4) = ((3D0-30D0*X**2+35D0*X**4))/8D0
        P(5) = (X*(15D0-70D0*X**2+63D0*X**4))/8D0
        P(6) = ((-5D0+105D0*X**2-315D0*X**4+231D0*X**6))/16D0
        P(7) = (X*(-35D0+315D0*X**2-693D0*X**4+429D0*X**6))/16D0
        P(8) = ((35D0-1260D0*X**2+6930D0*X**4-12012D0*X**6+6435D0*
     &  X**8))/128D0
        P(9) = (X*(315D0-4620D0*X**2+18018D0*X**4-25740D0*X**6+
     &  12155D0*X**8))/128D0
        P(10) = ((-63D0+46189D0*X**10+3465D0*X**2-30030D0*X**4+
     &  90090D0*X**6-109395D0*X**8))/256D0
        P(11) = (X*(-693D0+88179D0*X**10+15015D0*X**2-90090D0*X**4+
     &  218790D0*X**6-230945D0*X**8))/256D0
        P(12) = ((231D0-1939938D0*X**10+676039D0*X**12-18018D0*X**2+
     &  225225D0*X**4-1021020D0*X**6+2078505D0*X**8))/1024D0
        P(13) = (X*(3003D0-4056234D0*X**10+1300075D0*X**12-90090D0*
     &  X**2+765765D0*X**4-2771340D0*X**6+4849845D0*X**8))/1024D0
        P(14) = ((-429D0+22309287D0*X**10-16900975D0*X**12+
     &  5014575D0*X**14+45045D0*X**2-765765D0*X**4+4849845D0*X**6-
     &  14549535D0*X**8))/2048D0
        P(15) = (X*(-6435D0+50702925D0*X**10-35102025D0*X**12+
     &  9694845D0*X**14+255255D0*X**2-2909907D0*X**4+14549535D0*X**
     &  6-37182145D0*X**8))/2048D0
        P(16) = ((6435D0-1487285800D0*X**10+1825305300D0*X**12-
     &  1163381400D0*X**14+300540195D0*X**16-875160D0*X**2+
     &  19399380D0*X**4-162954792D0*X**6+669278610D0*X**8))/32768D0
        P(17) = (X*(109395D0-3650610600D0*X**10+4071834900D0*X**12-
     &  2404321560D0*X**14+583401555D0*X**16-5542680D0*X**2+
     &  81477396D0*X**4-535422888D0*X**6+1859107250D0*X**8))/32768D0
        P(18) = ((-12155D0+10039179150D0*X**10-17644617900D0*X**12+
     &  18032411700D0*X**14-9917826435D0*X**16+2268783825D0*X**18+
     &  2078505D0*X**2-58198140D0*X**4+624660036D0*X**6-
     &  3346393050D0*X**8))/65536D0
        P(19) = (X*(-230945D0+26466926850D0*X**10-42075627300D0*X**
     &  12+39671305740D0*X**14-20419054425D0*X**16+4418157975D0*X**
     &  18+14549535D0*X**2-267711444D0*X**4+2230928700D0*X**6-
     &  10039179150D0*X**8))/65536D0
        P(20) = ((46189D0-116454478140D0*X**10+273491577450D0*X**12-
     &  396713057400D0*X**14+347123925225D0*X**16-167890003050D0*X**
     &  18-9699690D0*X**2+34461632205D0*X**20+334639305D0*X**4-
     &  4461857400D0*X**6+30117537450D0*X**8))/262144D0

	elseif (x.gt.0.9d0) then

	xm1 = x-1d0

        P(0) = 1D0
        P(1) = 1D0+XM1
        P(2) = 1D0+3D0*XM1+(3D0*XM1**2)/2D0
        P(3) = 1D0+6D0*XM1+(15D0*XM1**2)/2D0+(5D0*XM1**3)/2D0
        P(4) = 1D0+10D0*XM1+(45D0*XM1**2)/2D0+(35D0*XM1**3)/2D0+
     &  (35D0*XM1**4)/8D0
        P(5) = 1D0+15D0*XM1+(105D0*XM1**2)/2D0+70D0*XM1**3+(315D0*
     &  XM1**4)/8D0+(63D0*XM1**5)/8D0
        P(6) = 1D0+21D0*XM1+105D0*XM1**2+210D0*XM1**3+(1575D0*XM1**
     &  4)/8D0+(693D0*XM1**5)/8D0+(231D0*XM1**6)/16D0
        P(7) = 1D0+28D0*XM1+189D0*XM1**2+525D0*XM1**3+(5775D0*XM1**
     &  4)/8D0+(2079D0*XM1**5)/4D0+(3003D0*XM1**6)/16D0+(429D0*XM1**
     &  7)/16D0
        P(8) = 1D0+36D0*XM1+315D0*XM1**2+1155D0*XM1**3+(17325D0*
     &  XM1**4)/8D0+(9009D0*XM1**5)/4D0+(21021D0*XM1**6)/16D0+
     &  (6435D0*XM1**7)/16D0+(6435D0*XM1**8)/128D0
        P(9) = 1D0+45D0*XM1+495D0*XM1**2+2310D0*XM1**3+(45045D0*
     &  XM1**4)/8D0+(63063D0*XM1**5)/8D0+(105105D0*XM1**6)/16D0+
     &  (6435D0*XM1**7)/2D0+(109395D0*XM1**8)/128D0+(12155D0*XM1**
     &  9)/128D0
        P(10) = 1D0+55D0*XM1+(46189D0*XM1**10)/256D0+(1485D0*XM1**
     &  2)/2D0+4290D0*XM1**3+(105105D0*XM1**4)/8D0+(189189D0*XM1**
     &  5)/8D0+(105105D0*XM1**6)/4D0+(36465D0*XM1**7)/2D0+(984555D0*
     &  XM1**8)/128D0+(230945D0*XM1**9)/128D0
        P(11) = 1D0+66D0*XM1+(969969D0*XM1**10)/256D0+(88179D0*XM1**
     &  11)/256D0+(2145D0*XM1**2)/2D0+(15015D0*XM1**3)/2D0+
     &  (225225D0*XM1**4)/8D0+63063D0*XM1**5+(357357D0*XM1**6)/4D0+
     &  (328185D0*XM1**7)/4D0+(6235515D0*XM1**8)/128D0+(1154725D0*
     &  XM1**9)/64D0
        P(12) = 1D0+78D0*XM1+(10669659D0*XM1**10)/256D0+(2028117D0*
     &  XM1**11)/256D0+(676039D0*XM1**12)/1024D0+(3003D0*XM1**2)/
     &  2D0+(25025D0*XM1**3)/2D0+(225225D0*XM1**4)/4D0+153153D0*
     &  XM1**5+(1072071D0*XM1**6)/4D0+(1247103D0*XM1**7)/4D0+
     &  (31177575D0*XM1**8)/128D0+(8083075D0*XM1**9)/64D0
        P(13) = 1D0+91D0*XM1+(81800719D0*XM1**10)/256D0+(6084351D0*
     &  XM1**11)/64D0+(16900975D0*XM1**12)/1024D0+(1300075D0*XM1**
     &  13)/1024D0+(4095D0*XM1**2)/2D0+20020D0*XM1**3+(425425D0*
     &  XM1**4)/4D0+(1378377D0*XM1**5)/4D0+(2909907D0*XM1**6)/4D0+
     &  (2078505D0*XM1**7)/2D0+(130945815D0*XM1**8)/128D0+
     &  (88913825D0*XM1**9)/128D0
        P(14) = 1D0+105D0*XM1+(245402157D0*XM1**10)/128D0+
     &  (50702925D0*XM1**11)/64D0+(219712675D0*XM1**12)/1024D0+
     &  (35102025D0*XM1**13)/1024D0+(5014575D0*XM1**14)/2048D0+
     &  2730D0*XM1**2+30940D0*XM1**3+(765765D0*XM1**4)/4D0+
     &  (2909907D0*XM1**5)/4D0+(14549535D0*XM1**6)/8D0+(6235515D0*
     &  XM1**7)/2D0+(480134655D0*XM1**8)/128D0+(409003595D0*XM1**9)/
     &  128D0
        P(15) = 1D0+120D0*XM1+(1227010785D0*XM1**10)/128D0+
     &  (659138025D0*XM1**11)/128D0+(1977414075D0*XM1**12)/1024D0+
     &  (245714175D0*XM1**13)/512D0+(145422675D0*XM1**14)/2048D0+
     &  (9694845D0*XM1**15)/2048D0+3570D0*XM1**2+46410D0*XM1**3+
     &  (1322685D0*XM1**4)/4D0+(2909907D0*XM1**5)/2D0+(33948915D0*
     &  XM1**6)/8D0+(68590665D0*XM1**7)/8D0+(1577585295D0*XM1**8)/
     &  128D0+(409003595D0*XM1**9)/32D0
        P(16) = 1D0+136D0*XM1+(5317046735D0*XM1**10)/128D0+
     &  (3559345335D0*XM1**11)/128D0+(13841898525D0*XM1**12)/1024D0+
     &  (2375237025D0*XM1**13)/512D0+(2181340125D0*XM1**14)/2048D0+
     &  (300540195D0*XM1**15)/2048D0+(300540195D0*XM1**16)/32768D0+
     &  4590D0*XM1**2+67830D0*XM1**3+(2204475D0*XM1**4)/4D0+
     &  (5555277D0*XM1**5)/2D0+(74687613D0*XM1**6)/8D0+(175287255D0*
     &  XM1**7)/8D0+(4732755885D0*XM1**8)/128D0+(1460727125D0*XM1**
     &  9)/32D0
        P(17) = 1D0+153D0*XM1+(20508608835D0*XM1**10)/128D0+
     &  (8305139115D0*XM1**11)/64D0+(80283011445D0*XM1**12)/1024D0+
     &  (35628555375D0*XM1**13)/1024D0+(22540514625D0*XM1**14)/
     &  2048D0+(300540195D0*XM1**15)/128D0+(9917826435D0*XM1**16)/
     &  32768D0+(583401555D0*XM1**17)/32768D0+5814D0*XM1**2+96900D0*
     &  XM1**3+(3561075D0*XM1**4)/4D0+(20369349D0*XM1**5)/4D0+
     &  (156165009D0*XM1**6)/8D0+(105172353D0*XM1**7)/2D0+
     &  (13146544125D0*XM1**8)/128D0+(18989452625D0*XM1**9)/128D0
        P(18) = 1D0+171D0*XM1+(143560261845D0*XM1**10)/256D0+
     &  (34407004905D0*XM1**11)/64D0+(401415057225D0*XM1**12)/
     &  1024D0+(220897043325D0*XM1**13)/1024D0+(22540514625D0*XM1**
     &  14)/256D0+(3305942145D0*XM1**15)/128D0+(168603049395D0*XM1**
     &  16)/32768D0+(20419054425D0*XM1**17)/32768D0+(2268783825D0*
     &  XM1**18)/65536D0+(14535D0*XM1**2)/2D0+135660D0*XM1**3+
     &  (5595975D0*XM1**4)/4D0+(36038079D0*XM1**5)/4D0+(156165009D0*
     &  XM1**6)/4D0+(239028075D0*XM1**7)/2D0+(34181014725D0*XM1**8)/
     &  128D0+(56968357875D0*XM1**9)/128D0
        P(19) = 1D0+190D0*XM1+(462583065945D0*XM1**10)/256D0+
     &  (516105073575D0*XM1**11)/256D0+(1777695253425D0*XM1**12)/
     &  1024D0+(73632347775D0*XM1**13)/64D0+(148767396525D0*XM1**
     &  14)/256D0+(56201016465D0*XM1**15)/256D0+(1967035576275D0*
     &  XM1**16)/32768D0+(183771489825D0*XM1**17)/16384D0+
     &  (83945001525D0*XM1**18)/65536D0+(4418157975D0*XM1**19)/
     &  65536D0+(17955D0*XM1**2)/2D0+(373065D0*XM1**3)/2D0+
     &  (8580495D0*XM1**4)/4D0+15444891D0*XM1**5+(300317325D0*XM1**
     &  6)/4D0+(1035788325D0*XM1**7)/4D0+(83898854325D0*XM1**8)/
     &  128D0+(79755701025D0*XM1**9)/64D0
        P(20) = 1D0+210D0*XM1+(1387749197835D0*XM1**10)/256D0+
     &  (1777695253425D0*XM1**11)/256D0+(1777695253425D0*XM1**12)/
     &  256D0+(347123925225D0*XM1**13)/64D0+(843015246975D0*XM1**
     &  14)/256D0+(393407115255D0*XM1**15)/256D0+(17703320186475D0*
     &  XM1**16)/32768D0+(2266515041175D0*XM1**17)/16384D0+
     &  (1594955028975D0*XM1**18)/65536D0+(172308161025D0*XM1**19)/
     &  65536D0+(21945D0*XM1**2)/2D0+(34461632205D0*XM1**20)/
     &  262144D0+(504735D0*XM1**3)/2D0+(25741485D0*XM1**4)/8D0+
     &  25741485D0*XM1**5+(557732175D0*XM1**6)/4D0+(2151252675D0*
     &  XM1**7)/4D0+(195763993425D0*XM1**8)/128D0+(210265029975D0*
     &  XM1**9)/64D0

	elseif (x.lt.-0.9d0) then

	xp1 = x+1d0

        P(0) = 1D0
        P(1) = -1D0+XP1
        P(2) = 1D0-3D0*XP1+(3D0*XP1**2)/2D0
        P(3) = -1D0+6D0*XP1-(15D0*XP1**2)/2D0+(5D0*XP1**3)/2D0
        P(4) = 1D0-10D0*XP1+(45D0*XP1**2)/2D0-(35D0*XP1**3)/2D0+
     &  (35D0*XP1**4)/8D0
        P(5) = -1D0+15D0*XP1-(105D0*XP1**2)/2D0+70D0*XP1**3-(315D0*
     &  XP1**4)/8D0+(63D0*XP1**5)/8D0
        P(6) = 1D0-21D0*XP1+105D0*XP1**2-210D0*XP1**3+(1575D0*XP1**
     &  4)/8D0-(693D0*XP1**5)/8D0+(231D0*XP1**6)/16D0
        P(7) = -1D0+28D0*XP1-189D0*XP1**2+525D0*XP1**3-(5775D0*XP1**
     &  4)/8D0+(2079D0*XP1**5)/4D0-(3003D0*XP1**6)/16D0+(429D0*XP1**
     &  7)/16D0
        P(8) = 1D0-36D0*XP1+315D0*XP1**2-1155D0*XP1**3+(17325D0*
     &  XP1**4)/8D0-(9009D0*XP1**5)/4D0+(21021D0*XP1**6)/16D0-
     &  (6435D0*XP1**7)/16D0+(6435D0*XP1**8)/128D0
        P(9) = -1D0+45D0*XP1-495D0*XP1**2+2310D0*XP1**3-(45045D0*
     &  XP1**4)/8D0+(63063D0*XP1**5)/8D0-(105105D0*XP1**6)/16D0+
     &  (6435D0*XP1**7)/2D0-(109395D0*XP1**8)/128D0+(12155D0*XP1**
     &  9)/128D0
        P(10) = 1D0-55D0*XP1+(46189D0*XP1**10)/256D0+(1485D0*XP1**
     &  2)/2D0-4290D0*XP1**3+(105105D0*XP1**4)/8D0-(189189D0*XP1**
     &  5)/8D0+(105105D0*XP1**6)/4D0-(36465D0*XP1**7)/2D0+(984555D0*
     &  XP1**8)/128D0-(230945D0*XP1**9)/128D0
        P(11) = -1D0+66D0*XP1-(969969D0*XP1**10)/256D0+(88179D0*
     &  XP1**11)/256D0-(2145D0*XP1**2)/2D0+(15015D0*XP1**3)/2D0-
     &  (225225D0*XP1**4)/8D0+63063D0*XP1**5-(357357D0*XP1**6)/4D0+
     &  (328185D0*XP1**7)/4D0-(6235515D0*XP1**8)/128D0+(1154725D0*
     &  XP1**9)/64D0
        P(12) = 1D0-78D0*XP1+(10669659D0*XP1**10)/256D0-(2028117D0*
     &  XP1**11)/256D0+(676039D0*XP1**12)/1024D0+(3003D0*XP1**2)/
     &  2D0-(25025D0*XP1**3)/2D0+(225225D0*XP1**4)/4D0-153153D0*
     &  XP1**5+(1072071D0*XP1**6)/4D0-(1247103D0*XP1**7)/4D0+
     &  (31177575D0*XP1**8)/128D0-(8083075D0*XP1**9)/64D0
        P(13) = -1D0+91D0*XP1-(81800719D0*XP1**10)/256D0+(6084351D0*
     &  XP1**11)/64D0-(16900975D0*XP1**12)/1024D0+(1300075D0*XP1**
     &  13)/1024D0-(4095D0*XP1**2)/2D0+20020D0*XP1**3-(425425D0*
     &  XP1**4)/4D0+(1378377D0*XP1**5)/4D0-(2909907D0*XP1**6)/4D0+
     &  (2078505D0*XP1**7)/2D0-(130945815D0*XP1**8)/128D0+
     &  (88913825D0*XP1**9)/128D0
        P(14) = 1D0-105D0*XP1+(245402157D0*XP1**10)/128D0-
     &  (50702925D0*XP1**11)/64D0+(219712675D0*XP1**12)/1024D0-
     &  (35102025D0*XP1**13)/1024D0+(5014575D0*XP1**14)/2048D0+
     &  2730D0*XP1**2-30940D0*XP1**3+(765765D0*XP1**4)/4D0-
     &  (2909907D0*XP1**5)/4D0+(14549535D0*XP1**6)/8D0-(6235515D0*
     &  XP1**7)/2D0+(480134655D0*XP1**8)/128D0-(409003595D0*XP1**9)/
     &  128D0
        P(15) = -1D0+120D0*XP1-(1227010785D0*XP1**10)/128D0+
     &  (659138025D0*XP1**11)/128D0-(1977414075D0*XP1**12)/1024D0+
     &  (245714175D0*XP1**13)/512D0-(145422675D0*XP1**14)/2048D0+
     &  (9694845D0*XP1**15)/2048D0-3570D0*XP1**2+46410D0*XP1**3-
     &  (1322685D0*XP1**4)/4D0+(2909907D0*XP1**5)/2D0-(33948915D0*
     &  XP1**6)/8D0+(68590665D0*XP1**7)/8D0-(1577585295D0*XP1**8)/
     &  128D0+(409003595D0*XP1**9)/32D0
        P(16) = 1D0-136D0*XP1+(5317046735D0*XP1**10)/128D0-
     &  (3559345335D0*XP1**11)/128D0+(13841898525D0*XP1**12)/1024D0-
     &  (2375237025D0*XP1**13)/512D0+(2181340125D0*XP1**14)/2048D0-
     &  (300540195D0*XP1**15)/2048D0+(300540195D0*XP1**16)/32768D0+
     &  4590D0*XP1**2-67830D0*XP1**3+(2204475D0*XP1**4)/4D0-
     &  (5555277D0*XP1**5)/2D0+(74687613D0*XP1**6)/8D0-(175287255D0*
     &  XP1**7)/8D0+(4732755885D0*XP1**8)/128D0-(1460727125D0*XP1**
     &  9)/32D0
        P(17) = -1D0+153D0*XP1-(20508608835D0*XP1**10)/128D0+
     &  (8305139115D0*XP1**11)/64D0-(80283011445D0*XP1**12)/1024D0+
     &  (35628555375D0*XP1**13)/1024D0-(22540514625D0*XP1**14)/
     &  2048D0+(300540195D0*XP1**15)/128D0-(9917826435D0*XP1**16)/
     &  32768D0+(583401555D0*XP1**17)/32768D0-5814D0*XP1**2+96900D0*
     &  XP1**3-(3561075D0*XP1**4)/4D0+(20369349D0*XP1**5)/4D0-
     &  (156165009D0*XP1**6)/8D0+(105172353D0*XP1**7)/2D0-
     &  (13146544125D0*XP1**8)/128D0+(18989452625D0*XP1**9)/128D0
        P(18) = 1D0-171D0*XP1+(143560261845D0*XP1**10)/256D0-
     &  (34407004905D0*XP1**11)/64D0+(401415057225D0*XP1**12)/
     &  1024D0-(220897043325D0*XP1**13)/1024D0+(22540514625D0*XP1**
     &  14)/256D0-(3305942145D0*XP1**15)/128D0+(168603049395D0*XP1**
     &  16)/32768D0-(20419054425D0*XP1**17)/32768D0+(2268783825D0*
     &  XP1**18)/65536D0+(14535D0*XP1**2)/2D0-135660D0*XP1**3+
     &  (5595975D0*XP1**4)/4D0-(36038079D0*XP1**5)/4D0+(156165009D0*
     &  XP1**6)/4D0-(239028075D0*XP1**7)/2D0+(34181014725D0*XP1**8)/
     &  128D0-(56968357875D0*XP1**9)/128D0
        P(19) = -1D0+190D0*XP1-(462583065945D0*XP1**10)/256D0+
     &  (516105073575D0*XP1**11)/256D0-(1777695253425D0*XP1**12)/
     &  1024D0+(73632347775D0*XP1**13)/64D0-(148767396525D0*XP1**
     &  14)/256D0+(56201016465D0*XP1**15)/256D0-(1967035576275D0*
     &  XP1**16)/32768D0+(183771489825D0*XP1**17)/16384D0-
     &  (83945001525D0*XP1**18)/65536D0+(4418157975D0*XP1**19)/
     &  65536D0-(17955D0*XP1**2)/2D0+(373065D0*XP1**3)/2D0-
     &  (8580495D0*XP1**4)/4D0+15444891D0*XP1**5-(300317325D0*XP1**
     &  6)/4D0+(1035788325D0*XP1**7)/4D0-(83898854325D0*XP1**8)/
     &  128D0+(79755701025D0*XP1**9)/64D0
        P(20) = 1D0-210D0*XP1+(1387749197835D0*XP1**10)/256D0-
     &  (1777695253425D0*XP1**11)/256D0+(1777695253425D0*XP1**12)/
     &  256D0-(347123925225D0*XP1**13)/64D0+(843015246975D0*XP1**
     &  14)/256D0-(393407115255D0*XP1**15)/256D0+(17703320186475D0*
     &  XP1**16)/32768D0-(2266515041175D0*XP1**17)/16384D0+
     &  (1594955028975D0*XP1**18)/65536D0-(172308161025D0*XP1**19)/
     &  65536D0+(21945D0*XP1**2)/2D0+(34461632205D0*XP1**20)/
     &  262144D0-(504735D0*XP1**3)/2D0+(25741485D0*XP1**4)/8D0-
     &  25741485D0*XP1**5+(557732175D0*XP1**6)/4D0-(2151252675D0*
     &  XP1**7)/4D0+(195763993425D0*XP1**8)/128D0-(210265029975D0*
     &  XP1**9)/64D0

	endif

	if (nl.gt.20) then
	  do 10 i=20,nl-1
	    p(i+1) = ( (2d0*i+1d0)*x*p(i) - i*p(i-1) )/(i+1d0)
10	  continue
	endif

	end

C***********************************************************************
C     GAUSS - INTEGRATION                                               
C                                                                       
C     DURCH AUFRUF VON STUETZ WERDEN DIE GAUSS-STUETZPUNKTE UND         
C     GEWICHTE DEFINIERT                                                
C***********************************************************************
      SUBROUTINE STUETZ                                                 
C***********************************************************************
      IMPLICIT REAL*8 (A-H,M,O-Z)                                       
      DIMENSION XG1(1),W1(1)                                            
      DIMENSION XG4(4),W4(4)                                            
      DIMENSION XG8(8),W8(8)                                            
      DIMENSION XG16(16),W16(16)                                        
      DIMENSION XG20(20),W20(20)                                        
      DIMENSION XGN(20,20),WN(20,20)                                    
      COMMON/STUET1/XG1,W1,N1                                           
      COMMON/STUET4/XG4,W4,N4                                           
      COMMON/STUET8/XG8,W8,N8                                           
      COMMON/STUE16/XG16,W16,N16                                        
      COMMON/STUE20/XG20,W20,N20                                        
      COMMON/STUETN/XGN,WN                                              
      N1=1                                                              
      XG1(1)=.577350269189626D0                                         
      W1(1)=1D0                                                         
      XGN(1,1)=.577350269189626D0                                       
      WN(1,1)=1D0                                                       
      N4=4                                                              
      XG4(1)=.18343464249565D0                                          
      XG4(2)=.525532409916329D0                                         
      XG4(3)=.796666477413627D0                                         
      XG4(4)=.960289856497536D0                                         
      W4(1)=.362683783378362D0                                          
      W4(2)=.313706645877887D0                                          
      W4(3)=.222381034453374D0                                          
      W4(4)=.101228536290376D0                                          
      DO 10 I=1,4                                                       
         XGN(4,I) = XG4(I)                                              
         WN(4,I) = W4(I)                                                
10    CONTINUE                                                          
      N8=8                                                              
      XG8(1)=.095012509837637440185D0                                   
      XG8(2)=.281603550779258913230D0                                   
      XG8(3)=.458016777657227386342D0                                   
      XG8(4)=.617876244402643748447D0                                   
      XG8(5)=.755404408355003033895D0                                   
      XG8(6)=.865631202387831743880D0                                   
      XG8(7)=.944575023073232576078D0                                   
      XG8(8)=.989400934991649932596D0                                   
      W8(1)=.189450610455068496285D0                                    
      W8(2)=.182603415044923588867D0                                    
      W8(3)=.169156519395002538189D0                                    
      W8(4)=.149595988816576732081D0                                    
      W8(5)=.124628971255533872052D0                                    
      W8(6)=.095158511682492784810D0                                    
      W8(7)=.062253523938647892863D0                                    
      W8(8)=.027152459411754094852D0                                    
      DO 20 I=1,8                                                       
         XGN(8,I) = XG8(I)                                              
         WN(8,I) = W8(I)                                                
20    CONTINUE                                                          
      N16=16                                                            
      XG16(1)=.048307665688D0                                           
      XG16(2)=.144471961583D0                                           
      XG16(3)=.239287362252D0                                           
      XG16(4)=.331868602282D0                                           
      XG16(5)=.421351276131D0                                           
      XG16(6)=.506899908932D0                                           
      XG16(7)=.587715757241D0                                           
      XG16(8)=.663044266930D0                                           
      XG16(9)=.732182118740D0                                           
      XG16(10)=.794483795968D0                                          
      XG16(11)=.849367613733D0                                          
      XG16(12)=.896321155766D0                                          
      XG16(13)=.934906075938D0                                          
      XG16(14)=.964762255588D0                                          
      XG16(15)=.985611511545D0                                          
      XG16(16)=.997263861849D0                                          
      W16(1)=.096540088515D0                                            
      W16(2)=.095638720079D0                                            
      W16(3)=.093844399081D0                                            
      W16(4)=.091173878696D0                                            
      W16(5)=.087652093004D0                                            
      W16(6)=.083311924227D0                                            
      W16(7)=.078193895787D0                                            
      W16(8)=.072345794109D0                                            
      W16(9)=.065822222776D0                                            
      W16(10)=.058684093479D0                                           
      W16(11)=.050998059262D0                                           
      W16(12)=.042835898022D0                                           
      W16(13)=.034273862913D0                                           
      W16(14)=.025392065309D0                                           
      W16(15)=.016274394731D0                                           
      W16(16)=.007018610009D0                                           
      DO 30 I=1,16                                                      
         XGN(16,I) = XG16(I)                                            
         WN(16,I) = W16(I)                                              
30    CONTINUE                                                          
      N20=20                                                            
      XG20(1)=.038772417506050821933D0                                  
      XG20(2)=.116084070675255208483D0                                  
      XG20(3)=.192697580701371099716D0                                  
      XG20(4)=.268152185007253681141D0                                  
      XG20(5)=.341994090825758473007D0                                  
      XG20(6)=.413779204371605001525D0                                  
      XG20(7)=.483075801686178712909D0                                  
      XG20(8)=.549467125095128202076D0                                  
      XG20(9)=.612553889667980237953D0                                  
      XG20(10)=.671956684614179548379D0                                 
      XG20(11)=.727318255189927103281D0                                 
      XG20(12)=.778305651426519387695D0                                 
      XG20(13)=.824612230833311663196D0                                 
      XG20(14)=.865959503212259503821D0                                 
      XG20(15)=.902098806968874296728D0                                 
      XG20(16)=.932812808278676533361D0                                 
      XG20(17)=.957916819213791655805D0                                 
      XG20(18)=.977259949983774262663D0                                 
      XG20(19)=.990726238699457006453D0                                 
      XG20(20)=.998237709710559200350D0                                 
      W20(1)=.077505947978424811264D0                                   
      W20(2)=.077039818164247965588D0                                   
      W20(3)=.076110361900626242372D0                                   
      W20(4)=.074723169057968264200D0                                   
      W20(5)=.072886582395804059061D0                                   
      W20(6)=.070611647391286779695D0                                   
      W20(7)=.067912045815233903826D0                                   
      W20(8)=.064804013456601038075D0                                   
      W20(9)=.061306242492928939167D0                                   
      W20(10)=.057439769099391551367D0                                  
      W20(11)=.053227846983936824355D0                                  
      W20(12)=.048695807635072232061D0                                  
      W20(13)=.043870908185673271992D0                                  
      W20(14)=.038782167974472017640D0                                  
      W20(15)=.033460195282547847393D0                                  
      W20(16)=.027937006980023401098D0                                  
      W20(17)=.022245849194166957262D0                                  
      W20(18)=.016421058381907888713D0                                  
      W20(19)=.010498284531152813615D0                                  
      W20(20)=.004521277098533191258D0                                  
      DO 40 I=1,20                                                      
         XGN(20,I) = XG20(I)                                            
         WN(20,I) = W20(I)                                              
40    CONTINUE                                                          
      END                                                               

************************************************************************
       	function dsoft(p)
************************************************************************
*       soft-photon correction factor 
*-----------------------------------------------------------------------
*       10.6.99 Stefan Dittmaier
************************************************************************
       	implicit real*8 (a-z)
        complex*16 sp(7,7),csp(7,7),cspen
        real*8  vp(7,7),p(7,0:3),q(6),m(6)
        integer i,j,dir(7)
	integer qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast

        common/rcoptions/qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast
        common/fstate/qf(4),mf(4),nc4f,dir
        common/prods/sp,csp,vp
        COMMON/IR/LAMBDA,LAMBDA2
        common/tcuts/deltas,deltac,deltags,deltagc
        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

	Iii(i)   =  log(IR)+2d0*log(m(i)/2d0/p(i,0))
	Iij(i,j) =  log(IR)*log(abs(vp(i,j))/m(i)/m(j))-pi**2/3d0
     &		   -log(m(i)/2d0/p(i,0))**2-log(m(j)/2d0/p(j,0))**2
     &		   -cspen(dcmplx(1d0-4d0*p(i,0)*p(j,0)/abs(vp(i,j))))

        call setprods(p,dir,6)
	IR = deltas**2*vp(1,2)/lambda2

	q(1) = -1d0	
	q(2) = -1d0	
	m(1) = ml(1)
	m(2) = ml(1)
	do 10 i=1,4
	  q(i+2) = qf(i)
	  m(i+2) = mf(i)
10	continue

	dsoft = 0d0
	do 20 i=1,6
	  if (q(i).ne.0d0) dsoft = dsoft - q(i)**2*Iii(i)
	do 20 j=1,i-1
	  if (q(i)*q(j).ne.0d0) 
     &		dsoft = dsoft - 2d0*(-1)**(i+j)*q(i)*q(j)*Iij(i,j)
20	continue

	dsoft = dsoft*alpha0/2d0/pi 

        end

************************************************************************
       	function dsoft1(p)
************************************************************************
*       part 1 of soft-photon correction factor 
*         log(lambda) and log(mf) terms exactly cancel virtual counterparts
*	  no log(deltas) terms !
*	--> use kinematics & Born cs of virtual corrections 
*-----------------------------------------------------------------------
*       10.6.99 Stefan Dittmaier
************************************************************************
       	implicit real*8 (a-z)
        complex*16 sp(7,7),csp(7,7)
        real*8  vp(7,7),p(7,0:3),q(6),m(6)
        integer i,j,dir(7)
	integer qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast

        common/rcoptions/qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast
        common/fstate/qf(4),mf(4),nc4f,dir
        common/prods/sp,csp,vp
        COMMON/IR/LAMBDA,LAMBDA2
        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

	L(p2,mi2,mj2) =  (log(mi2/p2)+log(mj2/p2)+2d0)*log(lambda2/p2)
     &		        -log(mi2/p2)**2/2d0+log(mi2/p2)/2d0
     &		        -log(mj2/p2)**2/2d0+log(mj2/p2)/2d0

        call setprods(p,dir,6)

	q(1) = -1d0	
	q(2) = -1d0	
	m(1) = ml(1)
	m(2) = ml(1)
	do 10 i=1,4
	  q(i+2) = qf(i)
	  m(i+2) = mf(i)
10	continue

	if (qsoft.eq.1) then
           constff =  1d0    -pi**2/3d0 
           constfi =  1d0    -pi**2/2d0 
           constif = -3d0/2d0+pi**2/6d0 
           constii =  3d0/2d0-pi**2/3d0 
	elseif (qsoft.eq.2) then
           constff =  -0.5d0
           constfi =  -0.5d0
           constif =  -0.5d0
           constii =  -0.5d0
        endif

	dsoft1 = 0d0
	do 20 i=2,6
	do 20 j=1,i-1
	  if (q(i)*q(j).eq.0d0) goto 20
	  if ((i.le.2).and.(j.le.2)) const = 2d0*constii+1d0
	  if ((i.gt.2).and.(j.le.2)) const = constif+constfi+1d0
	  if ((i.gt.2).and.(j.gt.2)) const = 2d0*constff+1d0
     	  dsoft1 = dsoft1 - (-1)**(i+j)*q(i)*q(j)
     &			*( L(abs(vp(i,j)),m(i)**2,m(j)**2) + const )
20	continue

	dsoft1 = dsoft1*alpha0/2d0/pi 

        end

************************************************************************
       	function dsoft2(p)
************************************************************************
*       part 2 of soft-photon correction factor 
*         no cancellations of singularities against virtual corrections
*	  full log(deltas) dependence
*	--> use kinematics & Born cs of hard-photonic corrections 
*-----------------------------------------------------------------------
*       10.6.99 Stefan Dittmaier
************************************************************************
       	implicit real*8 (a-z)
        complex*16 sp(7,7),csp(7,7),cspen
        real*8  vp(7,7),p(7,0:3),q(6),m(6)
        integer i,j,dir(7)
	integer qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast

        common/rcoptions/qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast
        common/fstate/qf(4),mf(4),nc4f,dir
        common/prods/sp,csp,vp
        COMMON/IR/LAMBDA,LAMBDA2
        common/tcuts/deltas,deltac,deltags,deltagc
        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

	Iij(i,j) =  
     &	     log(deltas*p(1,0)/p(i,0))*(log(abs(vp(i,j))/m(i)**2)-1d0)
     &	    +log(deltas*p(1,0)/p(j,0))*(log(abs(vp(i,j))/m(j)**2)-1d0)
     &	    -cspen(dcmplx(1d0-4d0*p(i,0)*p(j,0)/abs(vp(i,j))))
     &      -pi**2/3d0
     &	    -3d0/2d0*log(m(i)*m(j)/abs(vp(i,j)))
     &	    -log(abs(vp(i,j))/4d0/p(i,0)**2)**2/4d0
     &	    -log(abs(vp(i,j))/4d0/p(j,0)**2)**2/4d0

        call setprods(p,dir,6)

	q(1) = -1d0	
	q(2) = -1d0	
	m(1) = ml(1)
	m(2) = ml(1)
	do 10 i=1,4
	  q(i+2) = qf(i)
	  m(i+2) = mf(i)
10	continue

	if (qsoft.eq.1) then
           constff =  1d0    -pi**2/3d0 
           constfi =  1d0    -pi**2/2d0 
           constif = -3d0/2d0+pi**2/6d0 
           constii =  3d0/2d0-pi**2/3d0 
	elseif (qsoft.eq.2) then
           constff =  -0.5d0
           constfi =  -0.5d0
           constif =  -0.5d0
           constii =  -0.5d0
        endif

	dsoft2 = 0d0
	do 20 i=2,6
	do 20 j=1,i-1
	  if (q(i)*q(j).eq.0d0) goto 20
	  if ((i.le.2).and.(j.le.2)) const = 2d0*constii+1d0
	  if ((i.gt.2).and.(j.le.2)) const = constif+constfi+1d0
	  if ((i.gt.2).and.(j.gt.2)) const = 2d0*constff+1d0
          dsoft2 = dsoft2 - (-1)**(i+j)*q(i)*q(j)
     $                  *(2d0*Iij(i,j) - const)
20	continue

	dsoft2 = dsoft2*alpha0/2d0/pi 

        end

************************************************************************
       	function dsoftg(p)
************************************************************************
*       soft-gluon correction factor 
*-----------------------------------------------------------------------
*       10.6.99 Stefan Dittmaier
************************************************************************
       	implicit real*8 (a-z)
        complex*16 sp(7,7),csp(7,7),cspen
        real*8  vp(7,7),p(7,0:3),q(6),m(6)
        integer i,j,dir(7)
	integer qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast

        common/rcoptions/qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast
        common/fstate/qf(4),mf(4),nc4f,dir
        common/prods/sp,csp,vp
        COMMON/IR/LAMBDA,LAMBDA2
        common/tcuts/deltas,deltac,deltags,deltagc
        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

	Iii(i)   =  log(IR)+2d0*log(m(i)/2d0/p(i,0))
	Iij(i,j) =  log(IR)*log(abs(vp(i,j))/m(i)/m(j))-pi**2/3d0
     &		   -log(m(i)/2d0/p(i,0))**2-log(m(j)/2d0/p(j,0))**2
     &		   -cspen(dcmplx(1d0-4d0*p(i,0)*p(j,0)/abs(vp(i,j))))

        call setprods(p,dir,6)
	IR = deltas**2*vp(1,2)/lambda2

	do 10 i=1,4
	  q(i+2) = qf(i)
	  m(i+2) = mf(i)
10	continue

	dsoftg = 0d0
        if ((q(3).ne.0d0).and.(q(3).ne.-1d0).and.
     &      (q(4).ne.0d0).and.(q(4).ne.-1d0))
     &	    dsoftg = dsoftg - Iii(3) - Iii(4) + 2d0*Iij(3,4)
        if ((q(5).ne.0d0).and.(q(5).ne.-1d0).and.
     &      (q(6).ne.0d0).and.(q(6).ne.-1d0))
     &	    dsoftg = dsoftg - Iii(5) - Iii(6) + 2d0*Iij(5,6)

	dsoftg = dsoftg*alphas*2d0/3d0/pi

        end

************************************************************************
       	function dsoftg1(p)
************************************************************************
*       part 1 of soft-gluon correction factor 
*         log(lambda) and log(mf) terms exactly cancel virtual counterparts
*	  no log(deltas) terms !
*	--> use kinematics & Born cs of virtual corrections 
*-----------------------------------------------------------------------
*       10.6.99 Stefan Dittmaier
************************************************************************
       	implicit real*8 (a-z)
        complex*16 sp(7,7),csp(7,7)
        real*8  vp(7,7),p(7,0:3),q(6),m(6)
        integer i,dir(7)
	integer qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast

        common/rcoptions/qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast
        common/fstate/qf(4),mf(4),nc4f,dir
        common/prods/sp,csp,vp
        COMMON/IR/LAMBDA,LAMBDA2
        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

	L(p2,mi2,mj2) =  (log(mi2/p2)+log(mj2/p2)+2d0)*log(lambda2/p2)
     &		        -log(mi2/p2)**2/2d0+log(mi2/p2)/2d0
     &		        -log(mj2/p2)**2/2d0+log(mj2/p2)/2d0

        call setprods(p,dir,6)

	do 10 i=1,4
	  q(i+2) = qf(i)
	  m(i+2) = mf(i)
10	continue

	if (qsoft.eq.1) then
           constff =  1d0-pi**2/3d0 
	elseif (qsoft.eq.2) then
           constff =  -0.5d0
        endif

	dsoftg1 = 0d0
	const = 2d0*(constff+1d0/2d0)

        if ((q(3).ne.0d0).and.(q(3).ne.-1d0).and.
     &      (q(4).ne.0d0).and.(q(4).ne.-1d0))
     &    dsoftg1 = dsoftg1 + L(abs(vp(3,4)),m(3)**2,m(4)**2) + const
        if ((q(5).ne.0d0).and.(q(5).ne.-1d0).and.
     &      (q(6).ne.0d0).and.(q(6).ne.-1d0))
     &    dsoftg1 = dsoftg1 + L(abs(vp(5,6)),m(5)**2,m(6)**2) + const

	dsoftg1 = dsoftg1*alphas*2d0/3d0/pi

        end

************************************************************************
       	function dsoftg2(p)
************************************************************************
*       part 2 of soft-gluon correction factor 
*         no cancellations of singularities against virtual corrections
*	  full log(deltas) dependence
*	--> use kinematics & Born cs of hard-photonic corrections 
*-----------------------------------------------------------------------
*       10.6.99 Stefan Dittmaier
************************************************************************
       	implicit real*8 (a-z)
        complex*16 sp(7,7),csp(7,7),cspen
        real*8  vp(7,7),p(7,0:3),q(6),m(6)
        integer i,j,dir(7)
	integer qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast

        common/rcoptions/qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast
        common/fstate/qf(4),mf(4),nc4f,dir
        common/prods/sp,csp,vp
        COMMON/IR/LAMBDA,LAMBDA2
        common/tcuts/deltas,deltac,deltags,deltagc
        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

	Iij(i,j) =  
     &	     log(deltas*p(1,0)/p(i,0))*(log(abs(vp(i,j))/m(i)**2)-1d0)
     &	    +log(deltas*p(1,0)/p(j,0))*(log(abs(vp(i,j))/m(j)**2)-1d0)
     &	    -cspen(dcmplx(1d0-4d0*p(i,0)*p(j,0)/abs(vp(i,j))))
     &      -pi**2/3d0
     &	    -3d0/2d0*log(m(i)*m(j)/abs(vp(i,j)))
     &	    -log(abs(vp(i,j))/4d0/p(i,0)**2)**2/4d0
     &	    -log(abs(vp(i,j))/4d0/p(j,0)**2)**2/4d0

        call setprods(p,dir,6)

	do 10 i=1,4
	  q(i+2) = qf(i)
	  m(i+2) = mf(i)
10	continue

	if (qsoft.eq.1) then
           constff =  1d0-pi**2/3d0 
	elseif (qsoft.eq.2) then
           constff =  -0.5d0
        endif

	dsoftg2 = 0d0
	const = 2d0*(constff+1d0/2d0)

        if ((q(3).ne.0d0).and.(q(3).ne.-1d0).and.
     &      (q(4).ne.0d0).and.(q(4).ne.-1d0))
     &		dsoftg2 = dsoftg2 + 2d0*Iij(3,4) - const
        if ((q(5).ne.0d0).and.(q(5).ne.-1d0).and.
     &      (q(6).ne.0d0).and.(q(6).ne.-1d0))
     &		dsoftg2 = dsoftg2 + 2d0*Iij(5,6) - const
	
	dsoftg2 = dsoftg2*alphas*2d0/3d0/pi

        end

***********************************************************************
        function dM2subend(p,pon)
************************************************************************
*       calculate contribution to |M|^2 
*	from endpoint part of dipole subtraction function
*	only photon contribution
*	Note: this is NOT the FULL endpoint part,
*             but only the contribution that is added to the virtual
*	      corrections in DPA !
*	      --> see A.Denner, S.Dittmaier, M.Roth, and D.Wackeroth, 
*		  hep-ph/0006307 
*-----------------------------------------------------------------------
*	qsoft = 1:  S.Dittmaier, Nucl. Phys. B565 (2000) 69.
*	qsoft = 2:  M.Roth, Dissertation, ETH Zurich, No. 13363, 1999.
*       qsoft = 3:  YFS-inspired 
*-----------------------------------------------------------------------
*       10.6.99 Stefan Dittmaier
*       changed 12.7.00 SD
************************************************************************
        implicit real*8 (a-z)
        complex*16 sp(7,7),csp(7,7)
        complex*16 mat(10,-1:1),matcmb(10,-1:1),matb(-1:1),fb(10,-1:1)
        real*8  vp(7,7),p(7,0:3),pon(7,0:3),q(6),m(6)
        integer i,j,dir(7)
	integer qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast

        common/rcoptions/qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast
        common/fstate/qf(4),mf(4),nc4f,dir
        common/sme/mat,matcmb,matb,fb,mbcmb
        common/prods/sp,csp,vp
        common/pol/pm,pp
        COMMON/IR/LAMBDA,LAMBDA2
        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

	L(p2,mi2,mj2) =  (log(mi2/p2)+log(mj2/p2)+2d0)*log(lambda2/p2)
     &		        -log(mi2/p2)**2/2d0+log(mi2/p2)/2d0
     &		        -log(mj2/p2)**2/2d0+log(mj2/p2)/2d0

	if (qsoft.eq.1) then
           constff =  1d0    -pi**2/3d0 
           constfi =  1d0    -pi**2/2d0 
           constif = -3d0/2d0+pi**2/6d0 
           constii =  3d0/2d0-pi**2/3d0 
	elseif (qsoft.eq.2) then
           constff =  -0.5d0
           constfi =  -0.5d0
           constif =  -0.5d0
           constii =  -0.5d0
        else if (qsoft.eq.3) then
           constff =  1d0/2d0-pi**2*2d0/3d0
           constfi =  1d0/2d0-pi**2/6d0
           constif =  1d0/2d0-pi**2/6d0
           constii =  1d0/2d0-pi**2*2d0/3d0
        endif

	call calcsme(p,pon)

	q(1) = -1d0	
	q(2) = -1d0	
	m(1) = ml(1)
	m(2) = ml(1)
	do 10 i=1,4
	  q(i+2) = qf(i)
	  m(i+2) = mf(i)
10	continue

	Gsub = 0d0
	do 20 i=2,6
	do 20 j=1,i-1
	  if (q(i)*q(j).eq.0d0) goto 20
	  if ((i.le.2).and.(j.le.2)) const = 2d0*constii+1d0
	  if ((i.gt.2).and.(j.le.2)) const = constif+constfi+1d0
	  if ((i.gt.2).and.(j.gt.2)) const = 2d0*constff+1d0
     	  Gsub = Gsub - (-1)**(i+j)*q(i)*q(j)
     &			*( L(abs(vp(i,j)),m(i)**2,m(j)**2) + const )
20	continue
 
        dM2subend = alpha0/2d0/pi*mbcmb*Gsub

        end

***********************************************************************
        function dM2subendg(p,pon)
************************************************************************
*       calculate contribution to |M|^2 
*	from endpoint part of dipole subtraction function
*	only gluon contribution
*	Note: this is NOT the FULL endpoint part,
*             but only the contribution that is added to the virtual
*	      corrections in DPA !
*	      --> see A.Denner, S.Dittmaier, M.Roth, and D.Wackeroth, 
*		  hep-ph/0006307 
*-----------------------------------------------------------------------
*	qsoft = 1:  S.Dittmaier, Nucl. Phys. B565 (2000) 69.
*	qsoft = 2:  M.Roth, Dissertation, ETH Zurich, No. 13363, 1999.
*       qsoft = 3:  YFS-inspired 
*-----------------------------------------------------------------------
*       10.6.99 Stefan Dittmaier
*       changed 12.7.00 SD
************************************************************************
        implicit real*8 (a-z)
        complex*16 sp(7,7),csp(7,7),cmw2
        complex*16 mat(10,-1:1),matcmb(10,-1:1),matb(-1:1),fb(10,-1:1)
        real*8  vp(7,7),p(7,0:3),pon(7,0:3),q(6),m(6)
        integer i,dir(7)
	integer qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast

        common/rcoptions/qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast
        common/fstate/qf(4),mf(4),nc4f,dir
        common/sme/mat,matcmb,matb,fb,mbcmb
        common/prods/sp,csp,vp
	common/offshell/kp2off,km2off,cmw2
        common/pol/pm,pp
        COMMON/IR/LAMBDA,LAMBDA2
        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

	L(p2,mi2,mj2) =  (log(mi2/p2)+log(mj2/p2)+2d0)*log(lambda2/p2)
     &		        -log(mi2/p2)**2/2d0+log(mi2/p2)/2d0
     &		        -log(mj2/p2)**2/2d0+log(mj2/p2)/2d0

	if (qsoft.eq.1) then
           constff =  1d0    -pi**2/3d0 
	elseif (qsoft.eq.2) then
           constff =  -0.5d0
        else if (qsoft.eq.3) then
           constff =  1d0/2d0-pi**2*2d0/3d0
        endif

	do 10 i=1,4
	  q(i+2) = qf(i)
	  m(i+2) = mf(i)
10	continue

        call setprods(p,dir,6)

	Gsubg = 0d0
	const = 2d0*(constff+1d0/2d0)
        if (qsoft.eq.3) const = 2d0*(constff+3d0/2d0)

	mbcmboffsh = M2_Born_offsh(p,0,0)

	if ((q(3).ne.0d0).and.(q(3).ne.-1d0).and.
     &	    (q(4).ne.0d0).and.(q(4).ne.-1d0)) 
     &    Gsubg = Gsubg + L(abs(vp(3,4)),m(3)**2,m(4)**2) + const 
	if ((q(5).ne.0d0).and.(q(5).ne.-1d0).and.
     &	    (q(6).ne.0d0).and.(q(6).ne.-1d0)) 
     &    Gsubg = Gsubg + L(abs(vp(5,6)),m(5)**2,m(6)**2) + const 

	dM2subendg = alphas*2d0/3d0/pi*mbcmboffsh*Gsubg

        end

************************************************************************
       	function dnonf(p,pon)
************************************************************************
*       relative correction factor 
*	for virtual non-factorizable corrections
*       	qnf = 1:  on-shell  Coulomb singularity
*       	qnf = 2:  off-shell Coulomb singularity
*-----------------------------------------------------------------------
*       11.6.99 Stefan Dittmaier
************************************************************************
       	implicit real*8 (a-z)
        complex*16 sp(7,7),csp(7,7),cmw2,kp,km
        complex*16 dC0_coul
        real*8  vp(7,7),p(7,0:3),pon(7,0:3),tt(2,4),ss(4,4)
        integer i,j,dir(7)

        common/fstate/qf(4),mf(4),nc4f,dir
        common/prods/sp,csp,vp
        COMMON/IR/LAMBDA,LAMBDA2
        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

        cmw2 = dcmplx(mw2,-mw*gw)

c*** off-shellness of W's from off-shell phase space
        call setprods(p,dir,6)
	kp2  = vp(3,4)
	km2  = vp(5,6)
	kp   = kp2-cmw2
	km   = km2-cmw2

c*** other scalar products from on-shell phase space
        call setprods(pon,dir,6)
	s    = vp(1,2)
        t    = vp(1,3)+vp(1,4)+vp(3,4)
        u    = vp(2,3)+vp(2,4)+vp(3,4)
	do 10 i=1,2
	do 10 j=1,4
	  tt(i,j)  = vp(i,j+2)
10	continue
	do 20 i=1,3
	do 20 j=i+1,4
	  ss(i,j)  = vp(i+2,j+2)
20	continue

c*** W-boson self-energy contributions
	dmm1 =  4d0-2d0*log(-kp/lambda/mw)-2d0*log(-km/lambda/mw)

c*** Coulomb diagram contribution
	dmm2 =  (2d0*mw2-s)*dC0_coul(s,kp2,km2)

 	dnonf = alpha/pi*(dmm1+dmm2)

c*** IM + IF + MF + MF' + FF' contributions
 	dnonf = dnonf - qf(1)*qf(3)*alpha/pi
     &	  *dnf23(kp,km,s,t,u,ss(2,3),ss(1,3),ss(2,4),ss(1,4),
     &			     tt(1,1),tt(2,1),tt(1,3),tt(2,3) ) 
 	dnonf = dnonf + qf(2)*qf(3)*alpha/pi
     &	  *dnf23(kp,km,s,t,u,ss(1,3),ss(2,3),ss(1,4),ss(2,4),
     &		     	     tt(1,2),tt(2,2),tt(1,3),tt(2,3) ) 
 	dnonf = dnonf - qf(2)*qf(4)*alpha/pi
     &	  *dnf23(kp,km,s,t,u,ss(1,4),ss(2,4),ss(1,3),ss(2,3),
     &			     tt(1,2),tt(2,2),tt(1,4),tt(2,4) ) 
 	dnonf = dnonf + qf(1)*qf(4)*alpha/pi
     &	  *dnf23(kp,km,s,t,u,ss(2,4),ss(1,4),ss(2,3),ss(1,3),
     &			     tt(1,1),tt(2,1),tt(1,4),tt(2,4) ) 

	dnonf = dnonf*alpha0/alpha

	end

************************************************************************
       	function dnf23(kp,km,s,t,u,s13,s23,s14,s24,t12,t22,t13,t23) 
************************************************************************
*	auxiliary function for virtual non-factorizable corrections
*-----------------------------------------------------------------------
*       26.7.99 Stefan Dittmaier
************************************************************************
       	implicit real*8 (a-z)
        complex*16 kp,km,dety2,dy,dy0,dy2,dy3,kkp,kkm,gf3,f2,f3,d00
        complex*16 ieps,betaw,xw,cspen,xspen,d0_wwf2f3

        COMMON/IR/LAMBDA,LAMBDA2
        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

	dety2(kkp,kkm,ss13,ss24) = 
     &	  - s23*(kkp*(mw2**2+ss13*ss24-s14*s23)+2d0*kkm*mw2*ss13)

        gf3(kkp,kkm,ss24)   =
     &    - 2d0*xspen(kkp/kkm,(s23+ss24)/mw2+ieps)
     &    + xspen(kkp/kkm,-xw)     - xspen(-mw2/(s23+ss24)+ieps,-xw)
     &    + xspen(kkp/kkm,-1d0/xw) - xspen(-mw2/(s23+ss24)+ieps,-1d0/xw)
     &    - cspen(-ss24/s23-ieps) + 2d0*log(-s23/mw2-ieps)*log(-kkm/mw2)
     &    - log(-(s23+ss24)/mw2-ieps)**2

        ieps  = dcmplx(0d0,1d-16)
        betaw = sqrt(1d0-4d0*mw2/s+ieps)
        xw    = (betaw-1d0)/(betaw+1d0)

        dy0 = -mw2**4-s13**2*s24**2-s14**2*s23**2
     &        +2d0*mw2**2*s13*s24+2d0*mw2**2*s14*s23+2d0*s13*s14*s23*s24 
        dy2 = dety2(kp,km,s13,s24)
        dy3 = dety2(km,kp,s24,s13)
        dy  = 2d0*s23*(kp*km*s14*s23-(kp*mw2+km*s13)*(km*mw2+kp*s24))

        f2  = gf3(km,kp,s13)
        f3  = gf3(kp,km,s24)
        d00 = d0_wwf2f3(s,s13,s23,s14,s24)

	dnf23 = - kp*km*s23*dy0/dy*d00
     &          - kp*dy3/dy*f3 - km*dy2/dy*f2 
     &          + log(lambda2/mw2)*log(-s23/mw2-ieps)
     &		+ 2d0*log(-kp/lambda/mw)*log((mw2-t)*t22/t12/(mw2-u))
     &		+ 2d0*log(-km/lambda/mw)*log((mw2-t)*t13/t23/(mw2-u))
     &  	+ cspen(dcmplx(1d0-(-mw2+t)/t23))
     &		+ cspen(dcmplx(1d0-(-mw2+t)/t12))
     &		- cspen(dcmplx(1d0-(-mw2+u)/t22))
     &		- cspen(dcmplx(1d0-(-mw2+u)/t13))

	end

************************************************************************
       	function dnonf2(p,pon)
************************************************************************
*       relative correction factor 
*	for virtual non-factorizable corrections
*       	qnf = 1:  on-shell  Coulomb singularity
*       	qnf = 2:  off-shell Coulomb singularity
*-----------------------------------------------------------------------
*       11.6.99 Stefan Dittmaier
************************************************************************
       	implicit real*8 (a-z)
        complex*16 sp(7,7),csp(7,7),cmw2,kp,km
        complex*16 dC0_coul,dC0_wf,D0_if,D0_wwf2a,E0_wwf2af3
        real*8  vp(7,7),p(7,0:3),pon(7,0:3),tt(2,4),ss(4,4)
        integer i,j,dir(7)

        common/fstate/qf(4),mf(4),nc4f,dir
        common/prods/sp,csp,vp
        COMMON/IR/LAMBDA,LAMBDA2
        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

        cmw2 = dcmplx(mw2,-mw*gw)

c*** off-shellness of W's from off-shell phase space
        call setprods(p,dir,6)
	kp2  = vp(3,4)
	km2  = vp(5,6)
	kp   = kp2-cmw2
	km   = km2-cmw2

c*** other scalar products from on-shell phase space
        call setprods(pon,dir,6)
	s    = vp(1,2)
        t    = vp(1,3)+vp(1,4)+vp(3,4)
        u    = vp(2,3)+vp(2,4)+vp(3,4)
	do 10 i=1,2
	do 10 j=1,4
	  tt(i,j)  = vp(i,j+2)
10	continue
	do 20 i=1,3
	do 20 j=i+1,4
	  ss(i,j)  = vp(i+2,j+2)
20	continue

c*** W-boson self-energy contributions
	dmm1 =  4d0-2d0*log(-kp/lambda/mw)-2d0*log(-km/lambda/mw)

c*** Coulomb diagram contribution
	dmm2 =  (2d0*mw2-s)*dC0_coul(s,kp2,km2)

c*** initial--intermediate contributions
	dim  = -(t-mw2)*( dC0_wf(kp,t,ml(1)) + dC0_wf(km,t,ml(1)) )
     &	       +(u-mw2)*( dC0_wf(kp,u,ml(1)) + dC0_wf(km,u,ml(1)) )

c*** intermediate--final C0 contributions
	dmf1  = 0d0
	if (qf(1).ne.0d0) dmf1 = dmf1 + qf(1)*mw2*dC0_wf(kp,0d0,mf(1))
	if (qf(2).ne.0d0) dmf1 = dmf1 - qf(2)*mw2*dC0_wf(kp,0d0,mf(2))
	if (qf(3).ne.0d0) dmf1 = dmf1 - qf(3)*mw2*dC0_wf(km,0d0,mf(3))
	if (qf(4).ne.0d0) dmf1 = dmf1 + qf(4)*mw2*dC0_wf(km,0d0,mf(4)) 

c*** initial--final contributions
	dif  = 0d0
	if (qf(1).ne.0d0) dif = dif 
     &		+ qf(1)*kp*( -tt(1,1)*D0_if(kp,t,tt(1,1),mf(1)) 
     &			     +tt(2,1)*D0_if(kp,u,tt(2,1),mf(1)) )
	if (qf(2).ne.0d0) dif = dif 
     &		+ qf(2)*kp*( +tt(1,2)*D0_if(kp,t,tt(1,2),mf(2)) 
     &			     -tt(2,2)*D0_if(kp,u,tt(2,2),mf(2)) )

	if (qf(3).ne.0d0) dif = dif 
     &		+ qf(3)*km*( -tt(1,3)*D0_if(km,u,tt(1,3),mf(3)) 
     &			     +tt(2,3)*D0_if(km,t,tt(2,3),mf(3)) )
	if (qf(4).ne.0d0) dif = dif 
     &		+ qf(4)*km*( +tt(1,4)*D0_if(km,u,tt(1,4),mf(4)) 
     &			     -tt(2,4)*D0_if(km,t,tt(2,4),mf(4)) )

c*** intermediate--final D0 contributions
	dmf2  = 0d0
	if (qf(1).ne.0d0) dmf2 = dmf2 - qf(1)*(ss(1,3)+ss(1,4))*kp
     &	    *D0_wwf2a(s,kp,km,ss(2,3),ss(1,3),ss(2,4),ss(1,4),mf(1)) 
	if (qf(2).ne.0d0) dmf2 = dmf2 + qf(2)*(ss(2,3)+ss(2,4))*kp
     &	    *D0_wwf2a(s,kp,km,ss(1,3),ss(2,3),ss(1,4),ss(2,4),mf(2)) 
	if (qf(3).ne.0d0) dmf2 = dmf2 + qf(3)*(ss(1,3)+ss(2,3))*km
     &	    *D0_wwf2a(s,km,kp,ss(2,4),ss(2,3),ss(1,4),ss(1,3),mf(3))  
	if (qf(4).ne.0d0) dmf2 = dmf2 - qf(4)*(ss(1,4)+ss(2,4))*km
     &	    *D0_wwf2a(s,km,kp,ss(1,3),ss(1,4),ss(2,3),ss(2,4),mf(4))

c*** final--final contributions
	dff = 0d0
        if (qf(1)*qf(3).ne.0d0) 
     &	  dff = dff + qf(1)*qf(3)*ss(1,3)*kp*km
     &		      *E0_wwf2af3(s,kp,km,
     &			ss(2,3),ss(1,3),ss(2,4),ss(1,4),mf(1),mf(3))
        if (qf(2)*qf(3).ne.0d0) 
     &	  dff = dff - qf(2)*qf(3)*ss(2,3)*kp*km
     &      	      *E0_wwf2af3(s,kp,km,
     &			ss(1,3),ss(2,3),ss(1,4),ss(2,4),mf(2),mf(3))
        if (qf(1)*qf(4).ne.0d0) 
     &	  dff = dff - qf(1)*qf(4)*ss(1,4)*kp*km
     &      	      *E0_wwf2af3(s,kp,km,
     &			ss(2,4),ss(1,4),ss(2,3),ss(1,3),mf(1),mf(4))
        if (qf(2)*qf(4).ne.0d0) 
     &	  dff = dff + qf(2)*qf(4)*ss(2,4)*kp*km
     &      	      *E0_wwf2af3(s,kp,km,
     &			ss(1,4),ss(2,4),ss(1,3),ss(2,3),mf(2),mf(4))

 	dnonf2 = alpha0/pi*(dmm1+dmm2+dim+dmf1+dif+dmf2+dff)

	end

**********************************************************************
        FUNCTION XSPEN(R,X)                                              
**********************************************************************
        COMPLEX*16 XSPEN,CSPEN,ETA,R,X
	xspen = cspen(1d0+r*x)+eta(-x,r)*log(1d0+r*x)
	end

************************************************************************
       	function dC0_coul(s,kp2,km2)
************************************************************************
*	difference (off-shell - on-shell) for Coulomb C0 function
*-----------------------------------------------------------------------
*       11.6.99 Stefan Dittmaier
************************************************************************
       	implicit real*8 (a-z)
        complex*16 dC0_coul,cspen,xspen
	complex*16 ieps,b,bw,xw,kp,km,cmw2
        integer qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast

        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)
        COMMON/IR/LAMBDA,LAMBDA2
        common/rcoptions/qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast

	ieps = dcmplx(0d0,1d-20)
        cmw2 = dcmplx(mw2,-mw*gw)
	kp   = kp2-cmw2
	km   = km2-cmw2
	d    = abs(kp2-km2)/s
        b    = sqrt(1d0-4d0*cmw2/s)
	bb   = sqrt((s-kp2-km2)**2-4d0*kp2*km2)/s
        bw   = sqrt(1d0-4d0*mw2/s+ieps)
        xw   = (bw-1d0)/(bw+1d0)
 
	dC0_coul = 1d0/s/bw*(
     &	   cspen(1d0-xw**2)-dcmplx(0d0,2d0*pi)*log(1d0-xw**2)
     &	  +log(-xw)**2+pi**2+2d0*log(xw)*log(kp/lambda/mw)
     &	  -xspen(km/kp,-xw)+xspen(km/kp,-1d0/xw)		)

	if (qnf.eq.2) 
     &	  dC0_coul = dC0_coul 
     &	   + dcmplx(0d0,2d0*pi)/s/bb*log((b-bb+d)/(b+bb+d))
     &	   - dcmplx(0d0,2d0*pi)/s/bw*log((kp+km+s*bw*d)/2d0/s/bw**2)

	end

************************************************************************
       	function dC0_wf(ck,t,m)
************************************************************************
*	difference (off-shell - on-shell) 
*	for intermediate-initial and intermediate-final C0 
*-----------------------------------------------------------------------
*       11.6.99 Stefan Dittmaier
************************************************************************
       	implicit real*8 (a-z)
        complex*16 dC0_wf,ck

        COMMON/IR/LAMBDA,LAMBDA2
        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

	if (t.le.0d0) then
	  dC0_wf = 1d0/(t-mw2)*( pi**2/6d0 + log(m*mw/(mw2-t))
     &		      *(log(ck/(t-mw2))+log(-ck/lambda2)+log(m/mw)) )
	else
	  write(*,*) 'dC0_wf with t>0 not implemented !'
	  stop
	endif

	end

************************************************************************
       	function D0_if(ck,t,tt,m) 
************************************************************************
*	initial-final D0
*-----------------------------------------------------------------------
*       11.6.99 Stefan Dittmaier
************************************************************************
       	implicit real*8 (a-z)
        complex*16 D0_if,ck,cspen

        COMMON/IR/LAMBDA,LAMBDA2
        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

	D0_if = ( 2d0*log(-tt/m/ml(1))*log(-ck/mw/lambda)
     &		 -log((mw2-t)/ml(1)/mw)**2-log(m/mw)**2
     &		 -pi**2/3d0-cspen(dcmplx(1d0-(t-mw2)/tt))   )/tt/ck

	end
 
**********************************************************************
        function D0_wwf2a(s,kp,km,s13,s23,s14,s24,m2)
**********************************************************************
*       Off-shell D0-function for photon exchange between
*       intermediate and final state (W- and f2)
*       D0(s,km2,m2^2,0,kp2,s234,MW^2-I*MW*GW,MW^2-I*MW*GW,0,m2^2)
*-----------------------------------------------------------------------
*       16.4.97 Stefan Dittmaier
**********************************************************************
        implicit real*8 (a-z)
        complex*16 d0_wwf2a,betaw,xw,kp,km,xspen,ieps

        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

        ieps  = dcmplx(0d0,1d-16)
        betaw = sqrt(1d0-4d0*mw2/s+ieps)
        xw    = (betaw-1d0)/(betaw+1d0)

        D0_wwf2a = -1d0/( kp*(s23+s24)+km*mw2 )*(
     &          +xspen(-mw2/(s23+s24)+ieps,-xw)
     &          +xspen(-mw2/(s23+s24)+ieps,-1d0/xw)
     &          -xspen(kp/km,-xw)
     &          -xspen(kp/km,-1d0/xw)
     &          +2d0*xspen(kp/km,(s23+s24)/mw2+ieps)
     &          +log(m2**2/mw2)*(log(kp/km)+log(-(s23+s24)/mw2-ieps)) )

	end

**********************************************************************
        function e0_wwf2af3(s,kp,km,s13,s23,s14,s24,m2,m3)
**********************************************************************
*       E0-function with photon exchange between f2 and f3
*-----------------------------------------------------------------------
*       16.4.97 Stefan Dittmaier
**********************************************************************
        implicit real*8 (a-z)
        complex*16 e0_wwf2af3,kp,km,d0_wwf2f3,d0_wwf2a,d0_wf2af3
        complex*16 d00,d01,d02,d03,d04,dy0,dy1,dy2,dy3,dy4,dy

        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

        dy  = -2d0*s23*( kp**2*s24*mw2+km**2*s13*mw2
     &                  +kp*km*(s13*s24-s14*s23+mw2**2) )
        dy0 = -(mw2**4+s13**2*s24**2+s14**2*s23**2
     &      -2d0*mw2**2*s13*s24-2d0*mw2**2*s14*s23-2d0*s13*s14*s23*s24)
        dy1 =  kp*( (s23+s24)*(s13*s24-s14*s23)+(s23-s24)*mw2**2 )
     &        +km*( 2d0*s13*s23+s13*s24+s14*s23-mw2**2)*mw2
        dy2 = -s23*( kp*(s13*s24-s14*s23+mw2**2)+2d0*km*s13*mw2 )
        dy3 = -s23*( km*(s13*s24-s14*s23+mw2**2)+2d0*kp*s24*mw2 )
        dy4 =  km*( (s13+s23)*(s13*s24-s14*s23)-(s13-s23)*mw2**2 )
     &        +kp*( s13*s24+s14*s23+2d0*s23*s24-mw2**2)*mw2

        d00 = d0_wwf2f3(s,s13,s23,s14,s24)
        d01 =  d0_wwf2a(s,kp,km,s13,s23,s14,s24,m2)
        d04 =  d0_wwf2a(s,km,kp,s24,s23,s14,s13,m3)
        d02 = d0_wf2af3(s,kp,s13,s23,s14,s24,m2,m3)
        d03 = d0_wf2af3(s,km,s24,s23,s14,s13,m3,m2)
        e0_wwf2af3 = ( dy0*d00+dy1*d01+dy2*d02+dy3*d03+dy4*d04 )/dy

        end

**********************************************************************
        function d0_wf2af3(s,kp,s13,s23,s14,s24,m2,m3)
**********************************************************************
*       D0-function for E0-function
*       (W-)-propagator contracted in E0-function
*       D0(m3^2,s123,0,m2^2,kp2,s23,0,m3^2,mw^2-i*gw*mw,m2^2)
*-----------------------------------------------------------------------
*       16.4.97 Stefan Dittmaier
**********************************************************************
        implicit real*8 (a-z)
        complex*16 d0_wf2af3,kp,cspen,ieps

        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)
        COMMON/IR/LAMBDA,LAMBDA2

        ieps = dcmplx(0d0,1d-16)
        d0_wf2af3 = 1d0/s23/kp*(
     &          -cspen(-s13/s23-ieps)-pi**2/3d0
     &          +2d0*log(-s23/m2/m3-ieps)*log(-kp/lambda/mw)
     &          -log(m2/mw)**2-log(-(s13+s23)/m3/mw-ieps)**2    )

	end

**********************************************************************
        function d0_wwf2f3(s,s13,s23,s14,s24)
**********************************************************************
*       D0-function for E0-function
*       photon propagator contracted in E0-function
*       D0(S,0,S23,0,S123,S234,MW,MW,0,0)
*-----------------------------------------------------------------------
*       16.4.97 Stefan Dittmaier
**********************************************************************
        implicit real*8 (a-z)
        complex*16 d0_wwf2f3,betaw,xw,kappa,y(2),xspen,ieps
	integer i

        common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,
     &               mxe,mxe2,ml(3),ml2(3),mqp(3),mqp2(3),mqm(3),mqm2(3)

        ieps  = dcmplx(0d0,1d-16)
        betaw = sqrt(1d0-4d0*mw2/s+ieps)
        xw    = (betaw-1d0)/(betaw+1d0)
        kappa = sqrt( mw2**4+s13**2*s24**2+s14**2*s23**2
     &                  -2d0*mw2**2*s13*s24-2d0*mw2**2*s14*s23
     &                  -2d0*s13*s14*s23*s24 - mw2**4*ieps)
        y(1)  = (mw2**2+s13*s24-s14*s23+kappa)/2d0/s13/mw2
        y(2)  = (mw2**2+s13*s24-s14*s23-kappa)/2d0/s13/mw2

        d0_wwf2f3 = 0d0
        do 100 i=1,2
          sgn = -3d0+2d0*i
          d0_wwf2f3 = d0_wwf2f3 + sgn/kappa*(
     &           xspen(-(s13+s23)/mw2-ieps,y(i))
     &          +xspen(-mw2/(s23+s24)+ieps,y(i))
     &          -xspen(xw,y(i))
     &          -xspen(1d0/xw,y(i))
     &          -log(1d0+s24/s23)*log(-y(i))
     &                   )
100     continue

	end
